"use strict";
var util = require("util");
var events = require("events");
var EventEmitter = events.EventEmitter || events;
var net = require("net");
var modbusSerialDebug = require("debug")("modbus-serial");

var crc16 = require("../utils/crc16");


/* TODO: const should be set once, maybe */
var EXCEPTION_LENGTH = 5;
var MIN_DATA_LENGTH = 6;
var MAX_BUFFER_LENGTH = 256;

var MODBUS_PORT = 502;

/**
 * Simulate a modbus-RTU port using TCP connection
 * @module TcpDirectRTUBufferPort
 *
 * @param {string} ip - ip address
 * @param {object} options - all options as JSON object
 * @constructor
 */
var TcpDirectRTUBufferPort = function(ip, options) {
    var modbus = this;
    modbus.ip = ip;
    modbus.openFlag = false;
    modbus.callback = null;
    modbus._transactionIdWrite = 1;

    // options
    if (typeof options === "undefined") options = {};
    modbus.port = options.port || MODBUS_PORT;

    // internal buffer
    modbus._buffer = new Buffer(0);

    // handle callback - call a callback function only once, for the first event
    // it will triger
    var handleCallback = function(had_error) {
        if (modbus.callback) {
            modbus.callback(had_error);
            modbus.callback = null;
        }
    };

    // create a socket
    modbus._client = new net.Socket();
    if (options.timeout) this._client.setTimeout(options.timeout);

    // register the port data event
    modbus._client.on("data", function onData(data) {
        // add data to buffer
        modbus._buffer = Buffer.concat([modbus._buffer, data]);

        // check if buffer include a complete modbus answer
        var expectedLength = modbus._length;
        var bufferLength = modbus._buffer.length;

        modbusSerialDebug({
            action: "receive serial rtu buffered port",
            data: data,
            buffer: modbus._buffer
        });

        // check data length
        if (expectedLength < MIN_DATA_LENGTH || bufferLength < EXCEPTION_LENGTH) return;

        // check buffer size for MAX_BUFFER_SIZE
        if (bufferLength > MAX_BUFFER_LENGTH) {
            modbus._buffer = modbus._buffer.slice(-MAX_BUFFER_LENGTH);
            bufferLength = MAX_BUFFER_LENGTH;
        }

        // loop and check length-sized buffer chunks
        var maxOffset = bufferLength - EXCEPTION_LENGTH;
        for (var i = 0; i <= maxOffset; i++) {
            var unitId = modbus._buffer[i];
            var functionCode = modbus._buffer[i + 1];

            if (unitId !== modbus._id) continue;

            if (functionCode === modbus._cmd && i + expectedLength <= bufferLength) {
                modbus._emitData(i, expectedLength);
                return;
            }
            if (functionCode === (0x80 | modbus._cmd) && i + EXCEPTION_LENGTH <= bufferLength) {
                modbus._emitData(i, EXCEPTION_LENGTH);
                return;
            }

            // frame header matches, but still missing bytes pending
            if (functionCode === (0x7f & modbus._cmd)) break;
        }
    });

    this._client.on("connect", function() {
        modbus.openFlag = true;
        handleCallback();
    });

    this._client.on("close", function(had_error) {
        modbus.openFlag = false;
        handleCallback(had_error);
        modbus.emit('close', modbus.ip + ':' + modbus.port)
    });

    this._client.on("error", function(had_error) {
        modbus.openFlag = false;
        handleCallback(had_error);
    });

    this._client.on("timeout", function() {
        modbus.openFlag = false;
        modbusSerialDebug("TcpDirectRTUBufferPort port: TimedOut");
        handleCallback(new Error("TcpDirectRTUBufferPort Connection Timed Out."));
        modbus.emit('timeout', modbus.ip + ':' + modbus.port)
    });

    /**
     * Check if port is open.
     *
     * @returns {boolean}
     */
    Object.defineProperty(this, "isOpen", {
        enumerable: true,
        get: function() {
            return this.openFlag;
        }
    });

    EventEmitter.call(this);
};
util.inherits(TcpDirectRTUBufferPort, EventEmitter);

/**
 * Emit the received response, cut the buffer and reset the internal vars.
 *
 * @param {number} start The start index of the response within the buffer.
 * @param {number} length The length of the response.
 * @private
 */
TcpDirectRTUBufferPort.prototype._emitData = function(start, length) {
    var buffer = this._buffer.slice(start, start + length);

    modbusSerialDebug({
        action: "emit data serial rtu buffered port",
        buffer: buffer
    });

    this.emit("data", buffer);
    this._buffer = this._buffer.slice(start + length);
};

/**
 * Simulate successful port open.
 *
 * @param callback
 */
TcpDirectRTUBufferPort.prototype.open = function(callback) {
    this.callback = callback;
    this._client.connect(this.port, this.ip);
};

/**
 * Simulate successful close port.
 *
 * @param callback
 */
TcpDirectRTUBufferPort.prototype.close = function(callback) {
    this.callback = callback;
    this._client.end();
};

/**
 * Send data to a modbus slave.
 *
 * @param {Buffer} data
 */
TcpDirectRTUBufferPort.prototype.write = function(data) {
    if (data.length < MIN_DATA_LENGTH) {
        modbusSerialDebug("expected length of data is to small - minimum is " + MIN_DATA_LENGTH);
        return;
    }

    var length = null;

    // remember current unit and command
    this._id = data[0];
    this._cmd = data[1];

    // calculate expected answer length
    switch (this._cmd) {
        case 1:
        case 2:
            length = data.readUInt16BE(4);
            this._length = 3 + parseInt((length - 1) / 8 + 1) + 2;
            break;
        case 3:
        case 4:
            length = data.readUInt16BE(4);
            this._length = 3 + 2 * length + 2;
            break;
        case 5:
        case 6:
        case 15:
        case 16:
            this._length = 6 + 2;
            break;
        default:
            // raise and error ?
            this._length = 0;
            break;
    }

    // send buffer to slave
    this._client.write(data);

    modbusSerialDebug({
        action: "send serial rtu buffered",
        data: data,
        unitid: this._id,
        functionCode: this._cmd
    });
};

/**
 * TCP RTU bufferd port for Modbus.
 *
 * @type {TcpRTUBufferedPort}
 */
module.exports = TcpDirectRTUBufferPort;